//-----------------------------------------------------------------------------
// File: Lighting.cpp
//
// Desc: Example code showing how to use D3D lights.
//
// Copyright (c) 1995-2001 Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------
{*----------------------------------------------------------------------------*}
{*  Direct3D sample from DirectX 8.1 SDK                                      *}
{*  Delphi adaptation by Alexey Barkovoy                                      *}
{*  E-Mail: clootie@reactor.ru                                                *}
{*                                                                            *}
{*  Modified: 19-Feb-2001                                                     *}
{*                                                                            *}
{*  Latest version can be downloaded from:                                    *}
{*     http://groups.yahoo.com/group/JEDI-DirectXExamples                     *}
{*     http://clootie.narod.ru                                                *}
{*----------------------------------------------------------------------------*}

{$I DirectX.inc}

unit LightingUnit;

interface

uses
  Windows, Messages,
  D3D8, D3DX8, D3DApp, D3DFont, D3DFile, D3DUtil, DXUtil;

//-----------------------------------------------------------------------------
// Defines, constants, and global variables
//-----------------------------------------------------------------------------

type
  // Custom D3D vertex format used by the vertex buffer
  PMyVertex = ^TMyVertex;
  TMyVertex = packed record
    p: TD3DXVector3;       // vertex position
    n: TD3DXVector3;       // vertex normal
  end;

  PMyVertexArray = ^TMyVertexArray;
  TMyVertexArray = array [0..MaxInt div SizeOf(TMyVertex) - 1] of TMyVertex;

const
  D3DFVF_MYVERTEX = (D3DFVF_XYZ or D3DFVF_NORMAL);

type
//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the
//       generic functionality needed in all Direct3D samples. CMyD3DApplication
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
  PMyD3DApplication = ^CMyD3DApplication;
  CMyD3DApplication = object(TD3DApplication)
  public
    m_pWallVB: IDirect3DVertexBuffer8; // Vertex buffer for the walls and floor
    m_pSphereMesh: ID3DXMesh;          // Representation of point light
    m_pConeMesh: ID3DXMesh;            // Representation of dir/spot light
    m_pFont: CD3DFont;                 // Font for drawing text
    m_light: TD3DLight8;               // Description of the D3D light
    m_n: LongWord;                     // Number of vertices in the ground grid along X
    m_m: LongWord;                      // Number of vertices in the ground grid along Z
    m_nTriangles: LongWord;             // Number of triangles in the ground grid

    function InitDeviceObjects: HResult; virtual;
    function RestoreDeviceObjects: HResult; virtual;
    function InvalidateDeviceObjects: HResult; virtual;
    function DeleteDeviceObjects: HResult; virtual;
    function FrameMove: HResult; virtual;
    function Render: HResult; virtual;
    function FinalCleanup: HResult; virtual;

  public
    constructor Create;
  end;

implementation

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create(0);

  m_strWindowTitle            := 'Lighting';
  m_bUseDepthBuffer           := FALSE;

  New(m_pFont,    Create('Arial', 12, D3DFONT_BOLD));
  m_pWallVB                   := nil;
  m_pSphereMesh               := nil;
  m_pConeMesh                 := nil;

  // Set up wall/floor mesh resolution.  Try changing m_n and m_m to see
  // how the lighting is affected.
  m_n := 32;
  m_m := 32;
  m_nTriangles := (m_n-1)*(m_m-1)*2;

end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects;
begin
  Result:= m_pFont^.InitDeviceObjects(m_pd3dDevice);
  if Failed(Result) then Exit;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects;
var
  v: PMyVertexArray;
  dX, dZ{, dU, dV}: Single;
  k, z, x: LongWord;
  mtrl: TD3DMaterial8;
  matIdentity: TD3DXMatrix;
  matView: TD3DXMatrix;
  vFromPt, vLookatPt, vUpVec: TD3DXVector3;
  matProj: TD3DXMatrix;
  fAspect: Single;
  light: TD3DLight8;                 // Description of the D3D light
begin
  // Create a square grid m_n*m_m for rendering the wall
  if FAILED(m_pd3dDevice.CreateVertexBuffer(m_nTriangles*3*SizeOf(TMyVertex),
                                            D3DUSAGE_WRITEONLY, D3DFVF_MYVERTEX,
                                            D3DPOOL_MANAGED, m_pWallVB)) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  // Fill in the grid vertex data
  m_pWallVB.Lock(0, 0, PByte(v), 0);
  dX := 1.0 / (m_n-1);
  dZ := 1.0 / (m_m-1);
//  dU := 1.0 / (m_n-1); //BAA: never used
//  dV := 1.0 / (m_m-1); //BAA: never used
  k := 0;
  for z:= 0 to (m_m - 2) do
  begin
    for x:= 0 to (m_n - 2) do
    begin
      v^[k].p  := D3DXVector3(10 * x*dX, 0.0, 10 * z*dZ);
      v^[k].n  := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v^[k].p  := D3DXVector3(10 * x*dX, 0.0, 10 * (z+1)*dZ);
      v^[k].n  := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v^[k].p  := D3DXVector3(10 * (x+1)*dX, 0.0, 10 * (z+1)*dZ);
      v^[k].n  := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v^[k].p  := D3DXVector3(10 * x*dX, 0.0, 10 * z*dZ);
      v^[k].n  := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v^[k].p  := D3DXVector3(10 * (x+1)*dX, 0.0, 10 * (z+1)*dZ);
      v^[k].n  := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v^[k].p  := D3DXVector3(10 * (x+1)*dX, 0.0, 10 * z*dZ);
      v^[k].n  := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
    end;
  end;
  m_pWallVB.Unlock;

  // Create sphere and cone meshes to represent the lights
  Result:= E_FAIL;
  if FAILED(D3DXCreateSphere(m_pd3dDevice, 0.25, 20, 20, m_pSphereMesh, nil)) then Exit;

  if FAILED(D3DXCreateCylinder(m_pd3dDevice, 0.0, 0.25, 0.5, 20, 20, m_pConeMesh, nil)) then Exit;

  // Set up a material
  D3DUtil_InitMaterial(mtrl, 1.0, 1.0, 1.0, 1.0);
  m_pd3dDevice.SetMaterial(mtrl);

  // Set miscellaneous render states
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE,   iFALSE);
  m_pd3dDevice.SetRenderState(D3DRS_SPECULARENABLE, iFALSE);

  // Set the world matrix
  D3DXMatrixIdentity(matIdentity);
  m_pd3dDevice.SetTransform(D3DTS_WORLD,  matIdentity);

  // Set the view matrix.
  vFromPt:=   D3DXVector3(-10,  10, -10);
  vLookatPt:= D3DXVector3(0.0, 0.0, 0.0);
  vUpVec:=    D3DXVector3(0.0, 1.0, 0.0);
  D3DXMatrixLookAtLH(matView, vFromPt, vLookatPt, vUpVec);
  m_pd3dDevice.SetTransform(D3DTS_VIEW, matView);

  // Set the projection matrix
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, 1.0, 100.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  // Turn on lighting.
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iTRUE);

  // Enable ambient lighting to a dim, grey light, so objects that
  // are not lit by the other lights are not completely black
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT,
    D3DCOLOR_COLORVALUE(0.25, 0.25, 0.25, 1.0));

  // Set light #0 to be a simple, faint grey directional light so
  // the walls and floor are slightly different shades of grey
  ZeroMemory(@light, SizeOf(light));
  light._Type := D3DLIGHT_DIRECTIONAL;
  light.Direction := D3DXVector3(0.3, -0.5, 0.2);
  with light.Diffuse do
  begin
    r := 0.25; g := 0.25; b := 0.25;
  end;
  m_pd3dDevice.SetLight(0, light);

  // Set light #1 to be a simple, bright directional light to use
  // on the mesh representing light #2
  ZeroMemory(@light, SizeOf(light));
  light._Type := D3DLIGHT_DIRECTIONAL;
  light.Direction := D3DXVector3(0.5, -0.5, 0.5);
  with light.Diffuse do
  begin
    r := 1.0; g := 1.0; b := 1.0;
  end;
  m_pd3dDevice.SetLight(1, light);

  // Light #2 will be the light used to light the floor and walls.  It will
  // be set up in FrameMove() since it changes every frame.

  // Restore the font
  m_pFont^.RestoreDeviceObjects;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove;
var
  dwCaps: DWORD;
  x, y, z: Single;
  v: TD3DXVector3;
begin
  ZeroMemory(@m_light, SizeOf(m_light));

  // Rotate through the various light types
  m_light._Type := TD3DLightType(1 + (Round(m_fTime) div 5) mod 3);

  // Make sure the light type is supported by the device.  If
  // D3DVTXPCAPS_POSITIONALLIGHTS is not set, the device does not support
  // point or spot lights, so change light #2's type to a directional light.
  dwCaps := m_d3dCaps.VertexProcessingCaps;
  if ((dwCaps and D3DVTXPCAPS_POSITIONALLIGHTS) = 0) then
  begin
    if (m_light._Type = D3DLIGHT_POINT) or (m_light._Type = D3DLIGHT_SPOT)
    then m_light._Type:= D3DLIGHT_DIRECTIONAL;
  end;

  // Values for the light position, direction, and color
  x := Sin(m_fTime * 2.000);
  y := Sin(m_fTime * 2.246);
  z := Sin(m_fTime * 2.640);

  m_light.Diffuse.r  := 0.5 + 0.5 * x;
  m_light.Diffuse.g  := 0.5 + 0.5 * y;
  m_light.Diffuse.b  := 0.5 + 0.5 * z;
  m_light.Range      := 100.0;

  case m_light._Type of
    D3DLIGHT_POINT:
    begin
      m_light.Position     := D3DXVec3Scale(v, D3DXVector3(x, y, z), 4.5)^;
      m_light.Attenuation1 := 0.4;
    end;
    D3DLIGHT_DIRECTIONAL:
      m_light.Direction    := D3DXVector3(x, y, z);
    D3DLIGHT_SPOT:
    begin
      m_light.Position     := D3DXVec3Scale(v, D3DXVector3(x, y, z), 2.0)^;
      m_light.Direction    := D3DXVector3(x, y, z);
      m_light.Theta        := 0.5;
      m_light.Phi          := 1.0;
      m_light.Falloff      := 1.0;
      m_light.Attenuation0 := 1.0;
    end;
  end;
  m_pd3dDevice.SetLight(2, m_light);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render;
var
  matWorld,
  matTrans,
  matRotate: TD3DXMatrix;

  vecFrom, vecAt, vecUp: TD3DXVector3;
  matWorldInv: TD3DXMatrix;

  strLight: String;
begin
  // Clear the viewport
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET, $000000ff, 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    m_pd3dDevice.SetStreamSource(0, m_pWallVB, SizeOf(TMyVertex));
    m_pd3dDevice.SetVertexShader(D3DFVF_MYVERTEX);

    // Turn on light #0 and #2, and turn off light #1
    m_pd3dDevice.LightEnable(0, TRUE);
    m_pd3dDevice.LightEnable(1, FALSE);
    m_pd3dDevice.LightEnable(2, TRUE);

    // Draw the floor
    D3DXMatrixTranslation(matTrans, -5.0, -5.0, -5.0);
    D3DXMatrixRotationZ(matRotate, 0.0);
    D3DXMatrixMultiply(matWorld, matRotate, matTrans);
    m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);
    m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLELIST, 0, m_nTriangles);

    // Draw the back wall
    D3DXMatrixTranslation(matTrans, 5.0, -5.0, -5.0);
    D3DXMatrixRotationZ(matRotate, D3DX_PI/2);
    D3DXMatrixMultiply(matWorld, matRotate, matTrans);
    m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);
    m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLELIST, 0, m_nTriangles);

    // Draw the side wall
    D3DXMatrixTranslation(matTrans, -5.0, -5.0, 5.0);
    D3DXMatrixRotationX(matRotate,  -D3DX_PI/2);
    D3DXMatrixMultiply(matWorld, matRotate, matTrans);
    m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);
    m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLELIST, 0, m_nTriangles);

    // Turn on light #1, and turn off light #0 and #2
    m_pd3dDevice.LightEnable(0, FALSE);
    m_pd3dDevice.LightEnable(1, TRUE);
    m_pd3dDevice.LightEnable(2, FALSE);

    // Draw the mesh representing the light
    if (m_light._Type = D3DLIGHT_POINT) then
    begin
      // Just position the point light -- no need to orient it
      D3DXMatrixTranslation(matWorld, m_light.Position.x,
        m_light.Position.y, m_light.Position.z);
      m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);
      m_pSphereMesh.DrawSubset(0);
    end else
    begin
      // Position the light and point it in the light's direction
      vecFrom:= D3DXVector3(m_light.Position.x, m_light.Position.y, m_light.Position.z );
      vecAt:= D3DXVector3(m_light.Position.x + m_light.Direction.x,
                          m_light.Position.y + m_light.Direction.y,
                          m_light.Position.z + m_light.Direction.z);
      vecUp:= D3DXVector3(0, 1, 0);
      D3DXMatrixLookAtLH(matWorldInv, vecFrom, vecAt, vecUp);
      D3DXMatrixInverse(matWorld, nil, matWorldInv);
      m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);
      m_pConeMesh.DrawSubset(0);
    end;

    // Output statistics
    m_pFont^.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats, 0);
    m_pFont^.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats, 0);
    case m_light._Type of
      D3DLIGHT_POINT: strLight:= 'Point Light' + #0;
      D3DLIGHT_SPOT:  strLight:= 'Spot Light' + #0;
     else
      strLight:= 'Directional Light' + #0;
    end;
    m_pFont^.DrawText(2, 40, D3DCOLOR_ARGB(255,255,255,255), PChar(@strLight[1]), 0);

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects;
begin
  SAFE_RELEASE(m_pWallVB);
  SAFE_RELEASE(m_pSphereMesh);
  SAFE_RELEASE(m_pConeMesh);
  m_pFont^.InvalidateDeviceObjects;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects;
begin
  m_pFont^.DeleteDeviceObjects;
  SAFE_RELEASE(m_pWallVB);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup;
begin
  // Cleanup D3D font
  SAFE_DELETE(m_pFont);

  Result:= S_OK;
end;

end.

